;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; lambda-v++ with levels, aspects and exceptions
;
; Extension to Éric Tanter's lambda-v++ language with levels and aspects (AOSD.09)
; last updated Dec 2nd, 2010 
; by Ismael Figueroa
; Exception modifications are marked with EH comments
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; lambda-v++ with levels and aspects
;
; version 1.1 - corresponds to AOSD.09 submission
; last updated Oct 12th, 2009
; by Éric Tanter
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
; differences with the paper semantics:
; - supports a more full-fledged language
; - all lambda expressions are evaluated to lambda values 
;   that are *tagged* with a reference id (used for equality tests in pointcuts)
;   and a level (-1 for all functions, except level-capturing functions, tagged
;   with the level at which they always run
;;

(module asplevels-ex-core scheme
  (require redex)
  (provide λal
           (rename-out (red λal-red)))
           
  
  (define-language λal
    ;; expressions
    (e (e e ...) (if e e e) x v
       (λ (x ...) e)
       (λs (x ...) e) ;; level-capturing lambda
       (let ((x e) ...) e) ; for convenience
       (begin e e ...)     ; .
       (up e) (down e) ; level shifting
       (try e with e)  ; EH: try/catch block
       (raise e)       ; EH: raise an exception carrying a value
       
       
       ;;---below: not user visible syntax ---;;
       (in-up e) (in-down e) ; 
       (in-shift l e)        ; 
       (jp j) (in-jp e)      ;
       (app/prim e e ...))
    
    ;; values and primitives
    (v (λ r l (x ...) e)    ;; each lambda is tagged with an id and its level of app (-1 for dyn)
       (list v ...) number #t #f 
       string prim unspecified)
    (prim  +  - list cons car cdr empty? eq? deploy)
    (r (variable-prefix r)) ;; to allow ref equality of functions
    
    (ex (exn l e)) ;; EH: a raised expression is tagged with its level
    
    ;; program context <l,J,A,E>
    (P (l J A E))
    (l number) ;; level
    
    ;; jp stack, jp, jp kind
    (J (stack j ...))
    (j (l k v v ...)) ;; kind, lambda, args
    (k "call") ;; here, only call
    
    ;; aspect environment 
    (A (asps a ...))
    (a (l v v)) ;; level, pc, adv
    
    ;; expr evaluation context
    (E (v ... E e ...)
       (if E e e)
       (begin E e e ...)
       (in-up E) (in-down E) (in-jp E) (in-shift l E)
       (app/prim v ... E e ...)
       (try E with e)  ; EH
       (try ex with E) ; EH
       (raise E)       ; EH     
       hole)
    (x variable-not-otherwise-mentioned))
  
  
  (define red
    (reduction-relation 
     λal
     
     ;; exceptions
     ; level capturing exception
     (--> (l_1 J A (in-hole E (raise v)))
          (l_1 J A (in-hole E (exn l_1 v)))
          "RaiseCapture")
     
     ; raise of raise: keep original level
     (--> (in-hole P (raise (exn l_2 v)))
          (in-hole P (exn l_2 v))
          "Reraise")
     
     ; no-exception
     (--> (in-hole P (try v with e))
          (in-hole P v)
          "TryV")
     
     ;; handler execution
     ; Non level-capturing function, but current level is the same as the exception level
     (--> (l_1 J A (in-hole E (try (exn l_ex v_ex) with (λ r l_h (x ...) e))))
          (l_1 J A (in-hole E ((λ r l_h (x ...) e) v_ex)))
          "HndEx1"
          (side-condition (and (equal? (term l_ex) (term l_1))
                               (equal? -1 (term l_h)))))
     
     ; Non level-capturing function and current level is not the same as the exception level
     (--> (l_1 J A (in-hole E (try (exn l_ex v_ex) with (λ r l_h (x ...) e))))
          (l_1 J A (in-hole E (exn l_ex v_ex)))
          "HndProp1"
          (side-condition (and (not (equal? (term l_ex) (term l_1)))
                               (equal? -1 (term l_h)))))
     
     ; Level capturing function is at the same level as the exception
     (--> (l_1 J A (in-hole E (try (exn l_ex v_ex) with (λ r l_h (x ...) e))))
          (l_h J A (in-hole E (in-shift l_1 ((λ r l_h (x ...) e) v_ex))))
          "HndEx2"
          (side-condition (and (equal? (term l_ex) (term l_h))
                               (not (equal? -1 (term l_h))))))
     
     ; Level capturing function is not at same level as the exception
     (--> (l_1 J A (in-hole E (try (exn l_ex v_ex) with (λ r l_h (x ...) e))))
          (l_1 J A (in-hole E (exn l_ex v_ex)))
          "HndProp2"
          (side-condition (and (not (equal? (term l_ex) (term l_h)))
                               (not (equal? -1 (term l_h))))))     
     
     ;; function tagging
     (--> (in-hole P (λ (x ...) e))
          (in-hole P (λ ,(gensym "r") -1  (x ...) e))
          "FunTag")
     
     ; level capturing case
     (--> (l_1 J A (in-hole E (λs (x ...) e)))
          (l_1 J A (in-hole E (λ ,(gensym "r") l_1 (x ...) e)))
          "FunTag-c")
     
     ;; arithmetic
     (--> (in-hole P (+ number_1 number_2))
          (in-hole P ,(+ (term number_1) (term number_2)))
          "+")
     (--> (in-hole P (- number_1 number_2))
          (in-hole P ,(- (term number_1) (term number_2)))
          "-")
     (--> (in-hole P (+ (exn l v) number_2))  ; EH propagation
          (in-hole P (exn l v))
          "add-exp-1")
     (--> (in-hole P (+ number_1 (exn l v)))  ; EH propagation
          (in-hole P (exn l v))
          "add-exp-2")
     (--> (in-hole P (- (exn l v) number_2))  ; EH propagation
          (in-hole P (exn l v))
          "sub-exp-1")
     (--> (in-hole P (- number_1 (exn l v)))  ; EH propagation
          (in-hole P (exn l v))
          "sub-exp-2")
     
     
     ;; if
     (--> (in-hole P (if #f e_1 e_2))
          (in-hole P e_2)
          "if-f")
     (--> (in-hole P (if v_1 e_1 e_2))
          (in-hole P e_1)
          "if-t"
          (side-condition (term v_1)))
     (--> (in-hole P (if (exn l v) e_1 e_2)) ; EH propagation
          (in-hole P (exn l v))
          "if-ex")
     
     ;; let (syntactic sugar)
     (--> (in-hole P (let ((x e) ...) e_0))
          (in-hole P (app/prim (λ (x ...) e_0) e ...))
          "let")
     
     ;; begin
     (--> (in-hole P (begin v e_1 e_2 ...))
          (in-hole P (begin e_1 e_2 ...))
          "seq")
     (--> (in-hole P (begin e))
          (in-hole P e)
          "outseq")
     (--> (in-hole P (begin (exn l v_2) e ...)) ; EH propagation
          (in-hole P (exn l v_2))
          "exseq")
     
     ;; eq?
     (--> (in-hole P (eq? v_1 v_2))
          (in-hole P #t)
          "eq-t"
          (side-condition (term (feq? v_1 v_2))))
     (--> (in-hole P (eq? v_1 v_2))
          (in-hole P #f)
          "eq-f"
          (side-condition (not (term (feq? v_1 v_2)))))
     (--> (in-hole P (eq? v_1 (exn l v_2))) ; EH propagation
          (in-hole P (exn l v_2))
          "eq-ex-1")
     (--> (in-hole P (eq? (exn l v_2) v_1)) ; EH propagation
          (in-hole P (exn l v_2))
          "eq-ex-2")
     
     ;; lists
     (--> (in-hole P (cons v_0 (list v_1 ...)))
          (in-hole P (list v_0 v_1 ...))
          "cons")
     (--> (in-hole P (car (list v_1 v_2 ...)))
          (in-hole P v_1)
          "car")
     (--> (in-hole P (cdr (list v_1 v_2 ...)))
          (in-hole P (list v_2 ...))
          "cdr")
     (--> (in-hole P (empty? (list)))
          (in-hole P #t)
          "empty?-t")
     (--> (in-hole P (empty? v_1))
          (in-hole P #f)
          "empty?-f"
          (side-condition (not (equal? (term v_1) (term (list))))))
     
     ;; shifting execution levels
     (--> (l_1 J A (in-hole E (up e)))  
          (,(add1 (term l_1)) J A (in-hole E (in-up e)))
          "InUp")
     (--> (l_1 J A (in-hole E (in-up v)))
          (,(sub1 (term l_1)) J A (in-hole E v))
          "OutUp")
     (--> (l_1 J A (in-hole E (down e)))
          (,(sub1 (term l_1)) J A (in-hole E (in-down e)))
          "InDwn")
     (--> (l_1 J A (in-hole E (in-down v)))
          (,(add1 (term l_1)) J A (in-hole E v))
          "OutDwn")
     
     ;; EH: Exception keeps its level when returns from up
     (--> (l_1 J A (in-hole E (in-up (exn l_2 v))))
          (,(sub1 (term l_1)) J A (in-hole E (exn l_2 v)))
          "OutUpEx")
     
     ;; EH: Exception keeps its level when returns from down
     (--> (l_1 J A (in-hole E (in-down (exn l_2 v))))
          (,(add1 (term l_1)) J A (in-hole E (exn l_2 v)))
          "OutDwnEx")
     
     ;; aspects
     ; deploy
     (--> (l_1 J (asps a ...) (in-hole E (deploy v_1 v_2)))
          (l_1 J (asps (,(add1 (term l_1)) v_1 v_2) a ...) (in-hole E unspecified))
          "Deploy")
     
     ; app/prim: beta red
     (--> (in-hole P (app/prim (λ r l (x ...) e) v ...))
          (in-hole P (subst-n (x v) ... e))
          "AppPrim"
          (side-condition (eq? -1 (term l))))
     
     ; app/prim for level-capturing function
     (--> (l_1 J A (in-hole E (app/prim (λ r l_2 (x ...) e) v ...)))
          (l_2 J A (in-hole E (in-shift l_1 (subst-n (x v) ... e))))
          "AppShift"
          (side-condition (not (eq? -1 (term l_2)))))
     
     ; EH propagation
     (--> (in-hole P ((λ r l (x ...) e) v_1 ... (exn l_e v_e) v ...))
          (in-hole P (exn l_e v_e))
          "AppEx")
     
     ; back from shift
     (--> (l_2 J A (in-hole E (in-shift l_1 v)))
          (l_1 J A (in-hole E v))
          "Shift")   
     
     ; EH  propagation
     (--> (l_2 J A (in-hole E (in-shift l_1 (exn l_ex v))))
          (l_1 J A (in-hole E (exn l_ex v)))
          "ShiftEx")
     
     ; app: call jp
     (--> (l_1 J A (in-hole E ((λ r l (x ...) e) v ...)))
          (l_1 J A (in-hole E (jp (,(add1 (term l_1)) "call" (λ r l (x ...) e) v ...))))
          "App")
     
     ; weaving:
     (--> (l_1 (stack j ...) (asps a ...) (in-hole E (jp (l_2 k (λ r l (x ...) e) v ...))))
          (l_1 (stack (l_2 k (λ r l (x ...) e) v ...) j ...) (asps a ...) 
               (in-hole E (in-jp (up (app/prim (W (l_2 k (λ r l (x ...) e) v ...)
                                                  (asps a ...)) 
                                               v ...)))))
          "Weave")

     (--> (l_1 (stack (l_2 k v_0 v ...) j ...) A (in-hole E (in-jp v_1)))
          (l_1 (stack j ...) A (in-hole E v_1))
          "OutJp")   
     
     ; EH propagation
     (--> (l_1 (stack (l_2 k v_0 v ...) j ...) A (in-hole E (in-jp (exn l_ex v_ex))))
          (l_1 (stack j ...) A (in-hole E (exn l_ex v_ex)))
          "OutJpEx")
     
     ))
  
  (define-metafunction λal
    W : j A -> e
    ;; generate woven advice, recurring on aspect environment
    ((W (l_1 k (λ r l (x ...) e) v ...) (asps (l_2 v_1 v_2) a ...))
     (app/prim (λ (p) ;;proceed
                 (if (eq? l_1 l_2) ;; level check 
                     (let ((c (v_1 (list k (λ r l (x ...) e) v ...))))
                       (if c 
                           (λ (x ...) (v_2 p c x ...))
                           p))
                     p))
               (W (l_1 k (λ r l (x ...) e) v ...) (asps a ...))))  ;; base case
    
    ;; empty aspect environment (base case) -> just proceed at level-1
    ((W (l_1 k (λ r l (x ...) e) v ...) (asps))
     ;; we capture the level of original computation, which was already incremented when
     ;; emiting the join point, so we use one level below that of the join point
     (λ ,(gensym "r") ,(sub1 (term l_1)) (x ...) (app/prim (λ r l (x ...) e) x ...))))
  
  
  ;; comparison function: uses eq? on all terms, except for 
  ;; λ values, uses reference equality
  (define-metafunction λal
    ((feq? (λ r_1 l_1 (x_1 ...) e_1) (λ r_1 l_2 (x_2 ...) e_2))
     #t)
    ((feq? (λ r_1 l_1 (x_1 ...) e_1) (λ r_2 l_2 (x_2 ...) e_2))
     #f
     (side-condition (not (eq? (term r_1) (term r_2)))))
    ((feq? any_1 any_2)
     ,(eq? (term any_1) (term any_2))))
  
  
  ;;;; substitution functions
  (define-metafunction λal
    subst-n : (x any) ... any -> any 
    [(subst-n (x_1 any_1) (x_2 any_2) ... any_3) 
     (subst x_1 any_1 (subst-n (x_2 any_2) ... any_3))] 
    [(subst-n any_3) any_3])
  
  (define-metafunction λal 
    subst : x any any -> any  
    ;; 1. x_1 bound, so don't continue in λ body  
    [(subst x_1 any_1 (λ r l (x_2 ... x_1 x_3 ...) any_2)) 
     (λ r l (x_2 ... x_1 x_3 ...) any_2) 
     (side-condition (not (member (term x_1) (term (x_2 ...)))))]  
    
    ;; 2. general purpose capture avoiding case  
    [(subst x_1 any_1 (λ r l (x_2 ...) any_2)) 
     ,(term-let ([(x_new ...) 
                  (variables-not-in (term (x_1 any_1 any_2)) 
                                    (term (x_2 ...)))]) 
                (term 
                 (λ r l (x_new ...) 
                   (subst x_1 any_1 (subst-vars (x_2 x_new) ... any_2)))))]  
    
    ;; repeated, for substituting in lambdas that are not values yet
    
    ;; 1'. x_1 bound, so don't continue in λ body  
    [(subst x_1 any_1 (λ (x_2 ... x_1 x_3 ...) any_2)) 
     (λ (x_2 ... x_1 x_3 ...) any_2) 
     (side-condition (not (member (term x_1) (term (x_2 ...)))))]  
    
    ;; 2'. general purpose capture avoiding case  
    [(subst x_1 any_1 (λ (x_2 ...) any_2)) 
     ,(term-let ([(x_new ...) 
                  (variables-not-in (term (x_1 any_1 any_2)) 
                                    (term (x_2 ...)))]) 
                (term 
                 (λ (x_new ...) 
                   (subst x_1 any_1 (subst-vars (x_2 x_new) ... any_2)))))]    

    ;; same as above for level capturing lambdas that are not values yet
    ;; 1'. x_1 bound, so don't continue in λ body  
    [(subst x_1 any_1 (λs (x_2 ... x_1 x_3 ...) any_2)) 
     (λs (x_2 ... x_1 x_3 ...) any_2) 
     (side-condition (not (member (term x_1) (term (x_2 ...)))))]  
    
    ;; 2'. general purpose capture avoiding case  
    [(subst x_1 any_1 (λs (x_2 ...) any_2)) 
     ,(term-let ([(x_new ...) 
                  (variables-not-in (term (x_1 any_1 any_2)) 
                                    (term (x_2 ...)))]) 
                (term 
                 (λs (x_new ...) 
                   (subst x_1 any_1 (subst-vars (x_2 x_new) ... any_2)))))]
    
    ;; 3. replace x_1 with e_1  
    [(subst x_1 any_1 x_1) any_1]  
    ;; 4. x_1 and x_2 are different, so don't replace  
    [(subst x_1 any_1 x_2) x_2]  
    ;; the last two cases cover all other expression forms  
    [(subst x_1 any_1 (any_2 ...)) 
     ((subst x_1 any_1 any_2) ...)] 
    [(subst x_1 any_1 any_2) any_2])
  
  (define-metafunction λal 
    subst-vars : (x any) ... any -> any
    [(subst-vars (x_1 any_1) x_1) any_1] 
    [(subst-vars (x_1 any_1) (any_2 ...)) ((subst-vars (x_1 any_1) any_2) ...)] 
    [(subst-vars (x_1 any_1) any_2) any_2] 
    [(subst-vars (x_1 any_1) (x_2 any_2) ... any_3) 
     (subst-vars (x_1 any_1) (subst-vars (x_2 any_2) ... any_3))] 
    [(subst-vars any) any]))



